/*
    rate  --  statistic traffic analyzer
    Copyright (C) 2003 Mateusz 'mteg' Golicz

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <sys/select.h>
#include <sys/time.h>
#include <signal.h>
#include <time.h>
#include <assert.h>
#include <pcap.h>
#include <arpa/inet.h>
#include "scout.h"

static struct scout_context _g;

static inline double get_rate(unsigned long pp, unsigned long p, double s, double div)
{
	double x = p - pp;
	x /= div;
	x /= s;
	return x;
}

int stream_cmp(const void *a, const void *b)
{
	const struct scout_stream *s1 = *((struct scout_stream**) a);
	const struct scout_stream *s2 = *((struct scout_stream**) b);
	
	return (s2->p - s2->p_p) - (s1->p - s1->p_p);
}

struct scout_stream *_current_stream = NULL;
int pid_cmp(const void *a, const void *b)
{
	int p1 = *((int*) a);
	int p2 = *((int*) b);
	
	return (_current_stream->es[p2].p - _current_stream->es[p2].p_p) -
	       (_current_stream->es[p1].p - _current_stream->es[p1].p_p);
}

void report(struct scout_context *g)
{
	struct scout_stream *ss;
	struct in_addr in;
	int i;
	struct timespec ts;
	double s;
	time_t t = time(NULL);
	
	int pid_order[8192];
	int k;
	
	struct timespec *t0;
	
	static unsigned char stream_types[] = " MvaVA  ";
	
	clock_gettime(CLOCK_MONOTONIC, &ts);
	
	if(g->avg)
		t0 = &g->startup_time;
	else
		t0 = &g->last_report;

	s =  ts.tv_nsec - t0->tv_nsec;
	s /= 1000000000.0; 
	s += ts.tv_sec - t0->tv_sec;
	
	
	memcpy(&g->last_report, &ts, sizeof(struct timespec));
	
	printf("\n%s", ctime(&t));
	qsort(g->streams, g->streamcount, sizeof(struct scout_stream*), stream_cmp);
	
	for(i = 0; i<g->streamcount; i++)
	{
		int j, z;
		
		ss = g->streams[i];
		
		if(g->avg) ss->p_p = 0;
		
		in.s_addr = ss->saddr;
		printf("### %15s:%-5d   ---> ", inet_ntoa(in), ntohs(ss->source));
		
		in.s_addr = ss->daddr;
		printf(" %15s:%-5d %6.2f Mb/s, errors = %d\n", inet_ntoa(in), ntohs(ss->dest),
				get_rate(ss->p_p, ss->p, s, 1000000.0 / 188.0 / 8.0), ss->cc_errors);
		
		ss->p_p = ss->p;
		for(j = 0, k = 0; j<MAX_PID; j++)
		{
			if(ss->es[j].p)
			{
				pid_order[k++] = j;
				if(g->avg) ss->es[j].p_p = 0;
				
			}
		}
		
		_current_stream = ss;
		qsort(pid_order, k, sizeof(int), pid_cmp);

		for(z = 0; z<k; z++)
		{
			int p = pid_order[z];
			
			if(ss->es[p].sid)
				printf("   <%04d> ", ss->es[p].sid);
			else
				printf("          ");

			printf("%4d [0x%4x] %c %5.0f kb/s%s",
				p, p,
				stream_types[ss->es[p].flags & FL_STREAM_TYPE_MASK],
				get_rate(ss->es[p].p_p, ss->es[p].p, s, 1000.0 / 188.0 / 8.0),
				z & 1 ? "\n" : "");
			
			ss->es[p].p_p = ss->es[p].p;
		}
		
		if(z & 1) printf("\n");
		printf("\n");
	}
}

struct scout_stream * stream_find(struct scout_context *g, u_int32_t s, u_int32_t d, u_int16_t sp, u_int16_t dp)
{
	unsigned int h = s + d + sp + dp;
	struct scout_stream *ss;
	int i;
	
	h = h % HASH_SIZE;
	
	for(ss = g->stream_hash[h]; ss; ss = ss->hash_next)
		if(ss->saddr == s && ss->daddr == d && ss->source == sp && ss->dest == dp) return(ss);

	if(g->streamcount >= MAX_STREAMS) return(NULL);
	
	assert((ss = (struct scout_stream*) malloc(sizeof(struct scout_stream))));
	memset(ss, 0, sizeof(struct scout_stream));
	
	ss->saddr = s;
	ss->daddr = d;
	ss->source = sp;
	ss->dest = dp;
	
	for(i = 0; i<MAX_PID; i++)
		ss->es[i].next_cc = 16;
	
	
	g->streams[g->streamcount++] = ss;
	
	ss->hash_next = g->stream_hash[h];
	g->stream_hash[h] = ss;
	
	return ss;
}

int header_len(unsigned char *hdr)
{
	int l = 4;
	if(hdr[3] & 0x20)
		l += hdr[5] + 1;
	
	return l;
}

void parse_pat(struct scout_stream *ss, unsigned char *pat)
{
	int b = header_len(pat);
	int l = 188 - b;
	int sl;
	
	pat += b;
	pat ++;	/*account for pointer field */
	l -= 1;

	if(pat[0] != 0) return;
	
	sl  = ((pat[1] & 0x0f) << 8) | pat[2];
	sl -= 4;	/* CRC32 */
	sl += 3;	/* Bytes before section_length */

	if(sl < l) l = sl;	/* Seems a very long PAT */
	
	pat += 8;
	l   -= 8;
	
	while(l > 0)
	{
		u_int16_t pn  =   (pat[0] << 8) | pat[1];
		u_int16_t pid =  ((pat[2] << 8) | pat[3]) & 0x1fff;
		
		pat += 4;
		l   -= 4;
		
		ss->es[pid].flags &= ~FL_STREAM_TYPE_MASK;
		ss->es[pid].flags |= FL_PMT;
	}
	
	
}

void parse_pmt(struct scout_stream *ss, unsigned char *pmt, int pid)
{
	int b = header_len(pmt);
	int l = 188 - b;
	int sl, ch, pil;
	
	pmt += b;
	pmt ++;	/*account for pointer field */
	l -= 1;

	if(pmt[0] != 2) return;
	
	sl  = ((pmt[1] & 0x0f) << 8) | pmt[2];
	sl -= 4;	/* CRC32 */
	sl += 3;	/* Bytes before section_length */
	
	ch = (pmt[3] << 8) | pmt[4];
	
	pil = ((pmt[10] & 0x0f) << 8) | pmt[11];
	
	
	pmt += 12 + pil;
	l   -= 12 + pil;
	
	ss->es[pid].sid = ch;
	
	while(l > 0)
	{
		int st = pmt[0];
		int f = 0;
		int pid = ((pmt[1] << 8) | pmt[2]) & 0x1fff;
		int eil = ((pmt[3] << 8) | pmt[4]) & 0x0fff;
		
		pmt += 5 + eil;
		l   -= 5 + eil;
		
//		printf("pid = %d\n", pid);
		
		ss->es[pid].flags &= ~FL_STREAM_TYPE_MASK;
		switch(st)
		{
			case 2:  f = FL_VIDEO; break;
			case 4:  f = FL_AUDIO; break;
			case 6:  f = FL_AC3;	break;
			case 0x1b: f = FL_HD; 	break;
		}
		ss->es[pid].flags |= f;
		

		ss->es[pid].sid = ch;
	}
}


inline static u_int64_t capo_fetch_timestamp(unsigned char *rcvbuff)
{
	u_int64_t ts = 0;

	ts |= (rcvbuff[0] >> 1) & 0x7;		ts <<= 8;
	ts |= rcvbuff[1];			ts <<= 7;
	ts |= (rcvbuff[2] & 0xfe) >> 1;		ts <<= 8;
	ts |= rcvbuff[3];			ts <<= 7;
	ts |= (rcvbuff[4] & 0xfe) >> 1;
	return ts;
}

inline static void debug_search_pes(unsigned char *rcvbuff)
{
	int i;
	unsigned char *pes;
	int b = header_len(rcvbuff);
	int l = 188 - b;

	pes = rcvbuff + b;

	for(i = 0; i < l; i++)
	{
		if(pes[1] == 0 &&
		   pes[0] == 0 &&
		   pes[2] == 1 &&
		   pes[3] == 0xbd)	/* Private stream */
		{
			u_int64_t pts = 0, dts = 0;
			int phlen = pes[8] + 8;
						
			pts = capo_fetch_timestamp(rcvbuff + 9);
			dts = capo_fetch_timestamp(rcvbuff + 14);
						
			printf("### PSU (Offs %d Len %d), Stream %02x ",  i, (pes[4] << 8) | pes[5], pes[3]);

			if(pes[7] & 0x80)
				printf(" %llu ", pts);
			else
				printf(" [no pts]");

			if(pes[7] & 0x40)
				printf(" %llu ", dts);
			else
				printf(" [no dts] ");

			printf(" hdr len %d ", phlen);
			
			pes += phlen;
			
			printf(" di %02x did %02x ", pes[0], pes[1]);
			printf("\n");
			

			break;
		}
		rcvbuff++;
	}
}



void process_packet(unsigned char *userdata,const struct pcap_pkthdr *hd, const unsigned char *data)
{
	struct iphdr * iph;
	struct udphdr * udh;
	struct host *h;
	int bytes;
	
	struct scout_context * g = (struct scout_context*) userdata;
	unsigned char *mpeg_payload;
	
	struct scout_stream *ss;

	bytes = hd->len - g->prefix;
	
	g->total_packets++;
	g->total_bytes += bytes;


        iph = (struct iphdr*) (data + g->prefix);
	if(hd->len < (g->prefix + sizeof(struct iphdr))) return;
	if(iph->version != 4) return;



	if(iph->protocol  != IPPROTO_UDP)
	{
		g->non_udp++;
		return;
	}

	if(((iph->ihl * 4) + sizeof(struct udphdr)) > bytes) return;
	
	bytes -= iph->ihl * 4 + sizeof(struct udphdr);
	udh = (struct udphdr*) (data + g->prefix + iph->ihl * 4);
	mpeg_payload = (unsigned char*) (data + g->prefix + iph->ihl * 4 + sizeof(struct udphdr));
	
	if((bytes % 188) != 0)
	{
		g->non_mpeg++;
		return;
	}
	
	ss = stream_find(g, iph->saddr, iph->daddr, udh->source, udh->dest);
	if(!ss) return;
	
	for(; bytes >= 0; bytes -= 188, mpeg_payload += 188)
	{
		u_int16_t pid = ((mpeg_payload[1] & 0x1f) << 8) | mpeg_payload[2];
		int cc = mpeg_payload[3] & 0x0f;
		if(mpeg_payload[0] != 0x47) continue;

		ss->p++;
		ss->es[pid].p++;
		
		
		if(ss->es[pid].next_cc != cc && ss->es[pid].next_cc != 16 && (mpeg_payload[3] & 0x10))
			ss->cc_errors++;
		
		ss->es[pid].next_cc = (cc + 1) & 0xf;
		
		
		if(pid == 0)	parse_pat(ss, mpeg_payload);
		if((ss->es[pid].flags & FL_STREAM_TYPE_MASK) == FL_PMT)
			parse_pmt(ss, mpeg_payload, pid);
		
		if(pid == g->txt_pid && g->txt_pid)
			debug_search_pes(mpeg_payload);

	}
}


void generic_help(unsigned char * name)
{
	printf("mpeg scout - Mateusz 'mteg' Golicz <mtg@elsat.net.pl>, 2009 - 2010\n"
	       "usage: mpegscout <options>\n"
	       "  -p <pref> Datalink layer header size (gets substracted from each packet size, default: autosense or 14)\n"
	       "  -n <N>    Run only for N seconds\n"
	       "  -q        Skip regular reports, report only on exit\n"
	       "  -a        Show average flows\n"
	       "  -s        Group report by SIDs\n"
	       "  -v        Verbose output (use with -s)\n"
	       "  -W <PID>  VBI data PID - for interpreting WSS data\n"
	       "  -i <int>  Bind to interface <int> - default: eth0\n"
	      );
}

void alrm_callback(int s)
{
	_g.report = 1;
}

int main(int argc, char ** argv)
{
	pcap_t *handle;

	struct bpf_program bpf;
	char ebuff[PCAP_ERRBUF_SIZE];
	
	int c;
	
	struct itimerval itv;

	memset(&_g, 0, sizeof(struct scout_context));
	assert((_g.streams = (struct scout_stream**) malloc(sizeof(struct scout_stream*) * MAX_STREAMS)));
	
	

	_g.prefix = 14;
	_g.interface = "eth0";

	while((c = getopt(argc, argv, "p:i:hqn:asvW:")) != EOF)
	{
		switch(c)
		{
			case 'v': _g.v++; break;
			case 's': _g.sid = 1; break;
			case 'p': _g.prefix = atoi(optarg); break;
			case 'i': _g.interface = strdup(optarg); break;
			case 'h': generic_help(argv[0]); break;
			case 'a': _g.avg = 1; break;
			case 'q': _g.quiet = 1; break;
			case 'n': _g.cnt = atoi(optarg); break;
			case 'W': _g.txt_pid = atoi(optarg); break;
		}
	}

	if((handle = pcap_open_live(_g.interface, 65535, -1, 50, ebuff)) == NULL)
	{
		fprintf(stderr, "pcap_open_live(): %s\n", ebuff);
		return(0);
 	}

	signal(SIGALRM, alrm_callback);
	
	itv.it_interval.tv_sec = 1;
	itv.it_interval.tv_usec = 0;
	itv.it_value.tv_sec = 1;
	itv.it_value.tv_usec = 0;

	setitimer(ITIMER_REAL, &itv, NULL);
			
	setlinebuf(stdout);
		 
	clock_gettime(CLOCK_MONOTONIC, &_g.last_report);
	memcpy(&_g.startup_time, &_g.last_report, sizeof(struct timespec));
	
	while(1)
	{
		fd_set fds;
		int fd = pcap_get_selectable_fd(handle);
		struct timeval tv;
		
		FD_ZERO(&fds);
		FD_SET(fd, &fds);
		
		tv.tv_sec = 0;
		tv.tv_usec = 100000;
		
		if(select(fd + 1, &fds, NULL, NULL, &tv) < 0)
			if(errno != EINTR)
			{
				perror("select");
				break;
			}
		
		if(FD_ISSET(fd, &fds))
			pcap_dispatch(handle, -1, process_packet, &_g);
		
		if(_g.report)
		{
			if(!_g.quiet) report(&_g);
			_g.report = 0;
			
			if(_g.cnt)
			{
				_g.cnt--;
				if(!_g.cnt) break;
			}
		}
	}
	
	printf("--- Average\n");
	_g.avg = 0;
	report(&_g);
	
	pcap_close(handle);
	return 0;
}

